@@ -25,8 +25,6 @@ from oslo_db.sqlalchemy import utils as db_utils
 from oslo_log import log as logging
 from oslo_utils import uuidutils  # noqa
 import sqlalchemy as sa
-from sqlalchemy.ext.compiler import compiles
-from sqlalchemy.sql.expression import Insert
 
 from mistral import context
 from mistral.db.sqlalchemy import base as b
@@ -125,11 +123,6 @@ def refresh(model, session=None):
     session.refresh(model)
 
 
-@b.session_aware()
-def expire_all(session=None):
-    session.expire_all()
-
-
 @b.session_aware()
 def acquire_lock(model, id, session=None):
     # Expire all so all objects queried after lock is acquired
@@ -149,13 +142,16 @@ def _lock_entity(model, id):
 
 
 @b.session_aware()
-def update_on_match(id, specimen, values, session=None):
+def update_on_match(id, specimen, values, attempts, session=None):
     """Updates a model with the given values if it matches the given specimen.
 
     :param id: ID of a persistent model.
     :param specimen: Specimen used to match the
     :param values: Values to set to the model if fields of the object
         match the specimen.
+    :param attempts: The function will then invoke the UPDATE statement and
+        check for "success" one or more times, up to a maximum of that passed
+        as attempts.
     :param session: Session.
     :return: Persistent object attached to the session.
     """
@@ -180,7 +176,8 @@ def update_on_match(id, specimen, values, session=None):
         model = b.model_query(model_class).update_on_match(
             specimen=specimen,
             surrogate_key='id',
-            values=values
+            values=values,
+            attempts=attempts
         )
     except oslo_sqlalchemy.update_match.NoRowsMatched:
         LOG.info(
@@ -264,7 +261,7 @@ def _get_collection(model, insecure=False, limit=None, marker=None,
         if fields else ()
     )
 
-    query = (b.model_query(model, *columns) if insecure
+    query = (b.model_query(model, columns=columns) if insecure
              else _secure_query(model, *columns))
     query = db_filters.apply_filters(query, model, **filters)
 
@@ -280,30 +277,30 @@ def _get_collection(model, insecure=False, limit=None, marker=None,
     return query.all()
 
 
-def _get_db_object_by_name(model, name, filter_=None, order_by=None):
+def _get_db_object_by_name(model, name, columns=()):
+    query = _secure_query(model, *columns)
 
-    query = _secure_query(model)
-    final_filter = model.name == name
-
-    if filter_ is not None:
-        final_filter = sa.and_(final_filter, filter_)
-
-    if order_by is not None:
-        query = query.order_by(order_by)
-
-    return query.filter(final_filter).first()
+    return query.filter_by(name=name).first()
 
 
-def _get_db_object_by_id(model, id, insecure=False):
-    query = b.model_query(model) if insecure else _secure_query(model)
+def _get_db_object_by_id(model, id, insecure=False, columns=()):
+    query = (
+        b.model_query(model, columns=columns)
+        if insecure
+        else _secure_query(model, *columns)
+    )
 
     return query.filter_by(id=id).first()
 
 
 def _get_db_object_by_name_and_namespace_or_id(model, identifier,
-                                               namespace=None, insecure=False):
-
-    query = b.model_query(model) if insecure else _secure_query(model)
+                                               namespace=None, insecure=False,
+                                               columns=()):
+    query = (
+        b.model_query(model, columns=columns)
+        if insecure
+        else _secure_query(model, *columns)
+    )
 
     match_name = model.name == identifier
 
@@ -311,6 +308,7 @@ def _get_db_object_by_name_and_namespace_or_id(model, identifier,
         match_name = sa.and_(match_name, model.namespace == namespace)
 
     match_id = model.id == identifier
+
     query = query.filter(
         sa.or_(
             match_id,
@@ -321,42 +319,55 @@ def _get_db_object_by_name_and_namespace_or_id(model, identifier,
     return query.first()
 
 
-@compiles(Insert)
-def append_string(insert, compiler, **kw):
-    s = compiler.visit_insert(insert, **kw)
+def _get_db_object_by_name_and_namespace(model, name,
+                                         namespace, insecure=False,
+                                         columns=()):
+    query = (
+        b.model_query(model, columns=columns)
+        if insecure
+        else _secure_query(model, *columns)
+    )
 
-    if 'append_string' in insert.kwargs:
-        append = insert.kwargs['append_string']
+    if namespace is None:
+        namespace = ''
 
-        if append:
-            s += " " + append
+    query = query.filter(
+        sa.and_(
+            model.name == name,
+            model.namespace == namespace
+        )
+    )
 
-    if 'replace_string' in insert.kwargs:
-        replace = insert.kwargs['replace_string']
-
-        if isinstance(replace, tuple):
-            s = s.replace(replace[0], replace[1])
-
-    return s
+    return query.first()
 
 
 # Workbook definitions.
 
 @b.session_aware()
-def get_workbook(name, session=None):
-    wb = _get_db_object_by_name(models.Workbook, name)
+def get_workbook(name, namespace=None, fields=(), session=None):
+    wb = _get_db_object_by_name_and_namespace(
+        models.Workbook,
+        name,
+        namespace,
+        columns=fields
+    )
 
     if not wb:
         raise exc.DBEntityNotFoundError(
-            "Workbook not found [workbook_name=%s]" % name
+            "Workbook not found [name=%s, namespace=%s]" % (name, namespace)
         )
 
     return wb
 
 
 @b.session_aware()
-def load_workbook(name, session=None):
-    return _get_db_object_by_name(models.Workbook, name)
+def load_workbook(name, namespace=None, fields=(), session=None):
+    return _get_db_object_by_name_and_namespace(
+        models.Workbook,
+        name,
+        namespace,
+        columns=fields
+    )
 
 
 @b.session_aware()
@@ -372,9 +383,11 @@ def create_workbook(values, session=None):
 
     try:
         wb.save(session=session)
-    except db_exc.DBDuplicateEntry as e:
+    except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for WorkbookDefinition: %s" % e.columns
+            "Duplicate entry for WorkbookDefinition "
+            "['name', 'namespace', 'project_id']: {}, {}, {}".format(
+                wb.name, wb.namespace, wb.project_id)
         )
 
     return wb
@@ -382,7 +395,8 @@ def create_workbook(values, session=None):
 
 @b.session_aware()
 def update_workbook(name, values, session=None):
-    wb = get_workbook(name)
+    namespace = values.get('namespace')
+    wb = get_workbook(name, namespace=namespace)
 
     wb.update(values.copy())
 
@@ -398,13 +412,20 @@ def create_or_update_workbook(name, values, session=None):
 
 
 @b.session_aware()
-def delete_workbook(name, session=None):
+def delete_workbook(name, namespace=None, session=None):
+    namespace = namespace or ''
+
     count = _secure_query(models.Workbook).filter(
-        models.Workbook.name == name).delete()
+        sa.and_(
+            models.Workbook.name == name,
+            models.Workbook.namespace == namespace
+        )
+    ).delete()
 
     if count == 0:
         raise exc.DBEntityNotFoundError(
-            "Workbook not found [workbook_name=%s]" % name
+            "Workbook not found [workbook_name=%s, namespace=%s]"
+            % (name, namespace)
         )
 
 
@@ -416,12 +437,14 @@ def delete_workbooks(session=None, **kwargs):
 # Workflow definitions.
 
 @b.session_aware()
-def get_workflow_definition(identifier, namespace='', session=None):
+def get_workflow_definition(identifier, namespace='', fields=(), session=None):
     """Gets workflow definition by name or uuid.
 
     :param identifier: Identifier could be in the format of plain string or
                        uuid.
     :param namespace: The namespace the workflow is in. Optional.
+    :param fields: Fields that need to be loaded. For example,
+        (WorkflowDefinition.name,)
     :return: Workflow definition.
     """
     ctx = context.ctx()
@@ -430,7 +453,8 @@ def get_workflow_definition(identifier, namespace='', session=None):
         models.WorkflowDefinition,
         identifier,
         namespace=namespace,
-        insecure=ctx.is_admin
+        insecure=ctx.is_admin,
+        columns=fields
     )
 
     if not wf_def:
@@ -443,8 +467,12 @@ def get_workflow_definition(identifier, namespace='', session=None):
 
 
 @b.session_aware()
-def get_workflow_definition_by_id(id, session=None):
-    wf_def = _get_db_object_by_id(models.WorkflowDefinition, id)
+def get_workflow_definition_by_id(id, fields=(), session=None):
+    wf_def = _get_db_object_by_id(
+        models.WorkflowDefinition,
+        id,
+        columns=fields
+    )
 
     if not wf_def:
         raise exc.DBEntityNotFoundError(
@@ -455,20 +483,22 @@ def get_workflow_definition_by_id(id, session=None):
 
 
 @b.session_aware()
-def load_workflow_definition(name, namespace='', session=None):
+def load_workflow_definition(name, namespace='', fields=(), session=None):
     model = models.WorkflowDefinition
 
-    filter_ = model.namespace.in_([namespace, ''])
+    query = _secure_query(model, *fields)
+
+    filter_ = sa.and_(
+        model.name == name, model.namespace.in_([namespace, ''])
+    )
 
     # Give priority to objects not in the default namespace.
     order_by = model.namespace.desc()
 
-    return _get_db_object_by_name(
-        model,
-        name,
-        filter_,
-        order_by
-    )
+    if order_by is not None:
+        query = query.order_by(order_by)
+
+    return query.filter(filter_).first()
 
 
 @b.session_aware()
@@ -492,16 +522,17 @@ def create_workflow_definition(values, session=None):
 
     try:
         wf_def.save(session=session)
-    except db_exc.DBDuplicateEntry as e:
+    except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for WorkflowDefinition: %s" % e.columns
-        )
-
+            "Duplicate entry for WorkflowDefinition ['name', 'namespace',"
+            " 'project_id']: {}, {}, {}".format(wf_def.name, wf_def.namespace,
+                                                wf_def.project_id))
     return wf_def
 
 
 @b.session_aware()
-def update_workflow_definition(identifier, values, namespace='', session=None):
+def update_workflow_definition(identifier, values, session=None):
+    namespace = values.get('namespace')
     wf_def = get_workflow_definition(identifier, namespace=namespace)
 
     m_dbutils.check_db_obj_access(wf_def)
@@ -514,8 +545,8 @@ def update_workflow_definition(identifier, values, namespace='', session=None):
             if c_t.project_id != wf_def.project_id:
                 raise exc.NotAllowedException(
                     "Can not update scope of workflow that has cron triggers "
-                    "associated in other tenants. [workflow_identifier=%s]" %
-                    identifier
+                    "associated in other tenants. [workflow_identifier=%s, "
+                    "namespace=%s]" % (identifier, namespace)
                 )
 
         # Check event triggers.
@@ -523,12 +554,13 @@ def update_workflow_definition(identifier, values, namespace='', session=None):
             insecure=True,
             workflow_id=wf_def.id
         )
+
         for e_t in event_triggers:
             if e_t.project_id != wf_def.project_id:
                 raise exc.NotAllowedException(
                     "Can not update scope of workflow that has event triggers "
-                    "associated in other tenants. [workflow_identifier=%s]" %
-                    identifier
+                    "associated in other tenants. [workflow_identifier=%s, "
+                    "namespace=%s]" % (identifier, namespace)
                 )
 
     wf_def.update(values.copy())
@@ -538,10 +570,13 @@ def update_workflow_definition(identifier, values, namespace='', session=None):
 
 @b.session_aware()
 def create_or_update_workflow_definition(name, values, session=None):
-    if not _get_db_object_by_name(models.WorkflowDefinition, name):
-        return create_workflow_definition(values)
-    else:
+    namespace = values.get('namespace')
+    if _get_db_object_by_name_and_namespace_or_id(
+            models.WorkflowDefinition,
+            name,
+            namespace=namespace):
         return update_workflow_definition(name, values)
+    return create_workflow_definition(values)
 
 
 @b.session_aware()
@@ -554,8 +589,12 @@ def delete_workflow_definition(identifier, namespace='', session=None):
     if cron_triggers:
         raise exc.DBError(
             "Can't delete workflow that has cron triggers associated. "
-            "[workflow_identifier=%s], [cron_trigger_id(s)=%s]" %
-            (identifier, ', '.join([t.id for t in cron_triggers]))
+            "[workflow_identifier=%s, namespace=%s], [cron_trigger_id(s)=%s]"
+            % (
+                identifier,
+                namespace,
+                ', '.join([t.id for t in cron_triggers])
+            )
         )
 
     event_triggers = get_event_triggers(insecure=True, workflow_id=wf_def.id)
@@ -581,8 +620,12 @@ def delete_workflow_definitions(session=None, **kwargs):
 # Action definitions.
 
 @b.session_aware()
-def get_action_definition_by_id(id, session=None):
-    action_def = _get_db_object_by_id(models.ActionDefinition, id)
+def get_action_definition_by_id(id, fields=(), session=None):
+    action_def = _get_db_object_by_id(
+        models.ActionDefinition,
+        id,
+        columns=fields
+    )
 
     if not action_def:
         raise exc.DBEntityNotFoundError(
@@ -593,10 +636,11 @@ def get_action_definition_by_id(id, session=None):
 
 
 @b.session_aware()
-def get_action_definition(identifier, session=None):
+def get_action_definition(identifier, fields=(), session=None):
     a_def = _get_db_object_by_name_and_namespace_or_id(
         models.ActionDefinition,
-        identifier
+        identifier,
+        columns=fields
     )
 
     if not a_def:
@@ -608,8 +652,12 @@ def get_action_definition(identifier, session=None):
 
 
 @b.session_aware()
-def load_action_definition(name, session=None):
-    return _get_db_object_by_name(models.ActionDefinition, name)
+def load_action_definition(name, fields=(), session=None):
+    return _get_db_object_by_name(
+        models.ActionDefinition,
+        name,
+        columns=fields
+    )
 
 
 @b.session_aware()
@@ -625,9 +673,10 @@ def create_action_definition(values, session=None):
 
     try:
         a_def.save(session=session)
-    except db_exc.DBDuplicateEntry as e:
+    except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for action %s: %s" % (a_def.name, e.columns)
+            "Duplicate entry for Action ['name', 'project_id']:"
+            " {}, {}".format(a_def.name, a_def.project_id)
         )
 
     return a_def
@@ -665,8 +714,9 @@ def delete_action_definitions(session=None, **kwargs):
 # Action executions.
 
 @b.session_aware()
-def get_action_execution(id, session=None):
-    a_ex = _get_db_object_by_id(models.ActionExecution, id)
+def get_action_execution(id, insecure=False, fields=(), session=None):
+    a_ex = _get_db_object_by_id(models.ActionExecution, id, insecure=insecure,
+                                columns=fields)
 
     if not a_ex:
         raise exc.DBEntityNotFoundError(
@@ -677,8 +727,8 @@ def get_action_execution(id, session=None):
 
 
 @b.session_aware()
-def load_action_execution(id, session=None):
-    return _get_db_object_by_id(models.ActionExecution, id)
+def load_action_execution(id, fields=(), session=None):
+    return _get_db_object_by_id(models.ActionExecution, id, columns=fields)
 
 
 @b.session_aware()
@@ -696,15 +746,15 @@ def create_action_execution(values, session=None):
         a_ex.save(session=session)
     except db_exc.DBDuplicateEntry as e:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for ActionExecution: %s" % e.columns
+            "Duplicate entry for ActionExecution ID: {}".format(e.value)
         )
 
     return a_ex
 
 
 @b.session_aware()
-def update_action_execution(id, values, session=None):
-    a_ex = get_action_execution(id)
+def update_action_execution(id, values, insecure=False, session=None):
+    a_ex = get_action_execution(id, insecure)
 
     a_ex.update(values.copy())
 
@@ -742,13 +792,14 @@ def _get_action_executions(**kwargs):
 # Workflow executions.
 
 @b.session_aware()
-def get_workflow_execution(id, session=None):
+def get_workflow_execution(id, fields=(), session=None):
     ctx = context.ctx()
 
     wf_ex = _get_db_object_by_id(
         models.WorkflowExecution,
         id,
-        insecure=ctx.is_admin
+        insecure=ctx.is_admin,
+        columns=fields
     )
 
     if not wf_ex:
@@ -760,8 +811,8 @@ def get_workflow_execution(id, session=None):
 
 
 @b.session_aware()
-def load_workflow_execution(id, session=None):
-    return _get_db_object_by_id(models.WorkflowExecution, id)
+def load_workflow_execution(id, fields=(), session=None):
+    return _get_db_object_by_id(models.WorkflowExecution, id, columns=fields)
 
 
 @b.session_aware()
@@ -779,7 +830,7 @@ def create_workflow_execution(values, session=None):
         wf_ex.save(session=session)
     except db_exc.DBDuplicateEntry as e:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for WorkflowExecution with ID {value} ".format(
+            "Duplicate entry for WorkflowExecution with ID: {value} ".format(
                 value=e.value
             )
         )
@@ -829,14 +880,14 @@ def delete_workflow_executions(session=None, **kwargs):
 def update_workflow_execution_state(id, cur_state, state):
     specimen = models.WorkflowExecution(id=id, state=cur_state)
 
-    return update_on_match(id, specimen, {'state': state})
+    return update_on_match(id, specimen, values={'state': state}, attempts=1)
 
 
 # Tasks executions.
 
 @b.session_aware()
-def get_task_execution(id, session=None):
-    task_ex = _get_db_object_by_id(models.TaskExecution, id)
+def get_task_execution(id, fields=(), session=None):
+    task_ex = _get_db_object_by_id(models.TaskExecution, id, columns=fields)
 
     if not task_ex:
         raise exc.DBEntityNotFoundError(
@@ -847,8 +898,8 @@ def get_task_execution(id, session=None):
 
 
 @b.session_aware()
-def load_task_execution(id, session=None):
-    return _get_db_object_by_id(models.TaskExecution, id)
+def load_task_execution(id, fields=(), session=None):
+    return _get_db_object_by_id(models.TaskExecution, id, columns=fields)
 
 
 @b.session_aware()
@@ -945,7 +996,7 @@ def create_task_execution(values, session=None):
         task_ex.save(session=session)
     except db_exc.DBDuplicateEntry as e:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for TaskExecution: %s" % e.columns
+            "Duplicate entry for TaskExecution ID: {}".format(e.value)
         )
 
     return task_ex
@@ -987,7 +1038,7 @@ def delete_task_executions(session=None, **kwargs):
 def update_task_execution_state(id, cur_state, state):
     specimen = models.TaskExecution(id=id, state=cur_state)
 
-    return update_on_match(id, specimen, {'state': state})
+    return update_on_match(id, specimen, values={'state': state}, attempts=1)
 
 
 # Delayed calls.
@@ -1001,7 +1052,7 @@ def create_delayed_call(values, session=None):
         delayed_call.save(session)
     except db_exc.DBDuplicateEntry as e:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for DelayedCall: %s" % e.columns
+            "Duplicate entry for DelayedCall ID: {}".format(e.value)
         )
 
     return delayed_call
@@ -1093,6 +1144,19 @@ def get_expired_executions(expiration_time, limit=None, columns=(),
     return query.all()
 
 
+@b.session_aware()
+def get_running_expired_sync_actions(expiration_time, session=None):
+    query = b.model_query(models.ActionExecution)
+
+    query = query.filter(
+        models.ActionExecution.last_heartbeat < expiration_time
+    )
+    query = query.filter_by(is_sync=True)
+    query = query.filter(models.ActionExecution.state == states.RUNNING)
+
+    return query.all()
+
+
 @b.session_aware()
 def get_superfluous_executions(max_finished_executions, limit=None, columns=(),
                                session=None):
@@ -1111,9 +1175,12 @@ def get_superfluous_executions(max_finished_executions, limit=None, columns=(),
 
 def _get_completed_root_executions_query(columns):
     query = b.model_query(models.WorkflowExecution, columns=columns)
+
     # Only WorkflowExecution that are not a child of other WorkflowExecution.
-    query = query.filter(models.WorkflowExecution.
-                         task_execution_id == sa.null())
+    query = query.filter(
+        models.WorkflowExecution.task_execution_id == sa.null()
+    )
+
     query = query.filter(
         models.WorkflowExecution.state.in_(
             [states.SUCCESS,
@@ -1121,6 +1188,7 @@ def _get_completed_root_executions_query(columns):
              states.CANCELLED]
         )
     )
+
     return query
 
 
@@ -1186,10 +1254,10 @@ def create_cron_trigger(values, session=None):
 
     try:
         cron_trigger.save(session=session)
-    except db_exc.DBDuplicateEntry as e:
+    except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for cron trigger %s: %s"
-            % (cron_trigger.name, e.columns)
+            "Duplicate entry for cron trigger ['name', 'project_id']: "
+            "{}, {}".format(cron_trigger.name, cron_trigger.project_id)
         )
     # TODO(nmakhotkin): Remove this 'except' after fixing
     # https://bugs.launchpad.net/oslo.db/+bug/1458583.
@@ -1299,9 +1367,10 @@ def create_environment(values, session=None):
 
     try:
         env.save(session=session)
-    except db_exc.DBDuplicateEntry as e:
+    except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for Environment: %s" % e.columns
+            "Duplicate entry for Environment ['name', 'project_id']:"
+            " {}, {}".format(env.name, env.project_id)
         )
 
     return env
@@ -1387,9 +1456,13 @@ def create_resource_member(values, session=None):
 
     try:
         res_member.save(session=session)
-    except db_exc.DBDuplicateEntry as e:
+    except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for ResourceMember: %s" % e.columns
+            "Duplicate entry for ResourceMember ['resource_id',"
+            " 'resource_type', 'member_id']: {}, {}, "
+            "{}".format(res_member.resource_id,
+                        res_member.resource_type,
+                        res_member.member_id)
         )
 
     return res_member
@@ -1532,11 +1605,15 @@ def create_event_trigger(values, session=None):
 
     try:
         event_trigger.save(session=session)
-    except db_exc.DBDuplicateEntry as e:
+    except db_exc.DBDuplicateEntry:
         raise exc.DBDuplicateEntryError(
-            "Duplicate entry for event trigger %s: %s"
-            % (event_trigger.id, e.columns)
-        )
+            "Duplicate entry for EventTrigger ['exchange', 'topic',"
+            " 'event', 'workflow_id', 'project_id']:"
+            " {}, {}, {}, {}, {}".format(event_trigger.exchange,
+                                         event_trigger.topic,
+                                         event_trigger.event,
+                                         event_trigger.workflow_id,
+                                         event_trigger.project_id))
     # TODO(nmakhotkin): Remove this 'except' after fixing
     # https://bugs.launchpad.net/oslo.db/+bug/1458583.
     except db_exc.DBError as e:
